"""
This method is not currently used
"""

def _solve_iterative_init(self, solver, **kwds):
    """This method first fixes params and makes C = Z to solve for S. Requires fixed delta and sigmas.
    following this, the parameters are freed to users bounds and full problem is solved with fixed variances.
    This function is only meant to be used as part of a full variance estimation strategy. Currently not
    implemented in any of the strategies, however it can be useful in future implementations to initialize.
    
    Not meant to be directly used by users

    Args:
        sigma_sq (dict): variances 
    
        tee (bool,optional): flag to tell the optimizer whether to stream output
        to the terminal or not
    
        profile_time (bool,optional): flag to tell pyomo to time the construction and solution of the model. 
        Default False
    
        subset_lambdas (array_like,optional): Set of wavelengths to used in initialization problem 
        (Weifeng paper). Default all wavelengths.
        
        solver_options (dict, optional): Solver options for IPOPT

    Returns:

        None

    """   
    solver_opts = kwds.pop('solver_options', dict())
    tee = kwds.pop('tee', True)
    set_A = kwds.pop('subset_lambdas', list())
    profile_time = kwds.pop('profile_time', False)
    sigmas_sq = kwds.pop('variances', dict())

    if not set_A:
        set_A = self._meas_lambdas
    
    keys = sigmas_sq.keys()
    for k in self.component_set:
        print(k)
        if k not in keys:
            sigmas_sq[k] = 0.0
    
    self._warn_if_D_negative()
    list_components = [k for k in self._mixture_components]
             
    print("Solving Initialization Problem with fixed parameters\n")
    original_bounds = dict()
    for v, k in self.model.P.items():
        low = value(self.model.P[v].lb)
        high = value(self.model.P[v].ub)
        original_bounds[v] = (low, high)
        ub = value(self.model.P[v])
        lb = ub
        self.model.P[v].setlb(lb)
        self.model.P[v].setub(ub)
   
    obj = 0.0
   
    for t in self._meas_times:
        for l in set_A:
            D_bar = sum(self.model.Z[t, k] * self.model.S[l, k] for k in self.component_set)
            obj += (self.model.D[t, l] - D_bar) ** 2
    
    self.model.init_objective = Objective(expr=obj)
    opt = SolverFactory(solver)
    for key, val in solver_opts.items():
        opt.options[key]=val
    solver_results = opt.solve(self.model,
                               tee=tee,
                               report_timing=profile_time)
    
    for k, v in self.model.P.items():
        print(k, v.value)
    
    for t in self._allmeas_times:
        for k in self._mixture_components:
            if k in sigmas_sq and sigmas_sq[k] > 0.0:
                self.model.C[t, k].value = np.random.normal(self.model.Z[t, k].value, sigmas_sq[k])
            else:
                self.model.C[t, k].value = self.model.Z[t, k].value
            
    self.model.del_component('init_objective')
    
    for v, k in self.model.P.items():
        ub = original_bounds[v][1]
        lb = original_bounds[v][0]
        self.model.P[v].setlb(lb)
        self.model.P[v].setub(ub)
        
    m = self.model

    m.D_bar = Var(m.meas_times,
                  m.meas_lambdas)
    
    def rule_D_bar(m, t, l):
        return m.D_bar[t, l] == sum(getattr(m, self.component_var)[t, k] * m.S[l, k] for k in self.component_set)
    
    m.D_bar_constraint = Constraint(m.meas_times,
                                    m.meas_lambdas,
                                    rule=rule_D_bar)

    # estimation
    def rule_objective(m):
        expr = 0
        for t in m.meas_times:
            for l in m.meas_lambdas:

                expr += (m.D[t, l] - m.D_bar[t, l]) ** 2 / (sigmas_sq['device'])

        #expr = spectra_objective(m)

        # This part is not even used????
        # second_term = 0.0
        # for t in m.meas_times:
        #     second_term += sum((m.C[t, k] - m.Z[t, k]) ** 2 / sigmas_sq[k] for k in list_components)
        return expr

    m.objective = Objective(rule=rule_objective) 
    
    opt = SolverFactory(solver)

    for key, val in solver_opts.items():
        opt.options[key]=val
    
    print("Solving problem with unfixed parameters")
    solver_results = opt.solve(self.model,
                               tee=tee,
                               report_timing=profile_time)
    
    for k, v in self.model.P.items():
        print(k, v.value)      

    m.del_component('objective')
    self.model = m


"""
Data Tools left here for legacy reasons. Don't want to break the API do we!
"""
from kipet.common.read_write_tools import *
from kipet.common.diagnostic_tools import *
from kipet.common.prob_gen_tools import *


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec 10 11:02:58 2020

@author: kevin
"""
import matplotlib.pyplot as plt
import numpy as np
from pyomo.environ import *

from kipet.nsd_funs.NSD_TrustRegion_Ipopt import NSD


""" Model 1 """

m = ConcreteModel()

x_data = dict(zip(list(range(1, 6)), list(range(1, 6))))
y_data = dict(zip(list(range(1, 6)), [2, 4, 6, 8, 10]))

m.indx = Set(initialize=list(range(1, 6)))

m.X = Param(m.indx, initialize=x_data)
m.Y = Param(m.indx, initialize=y_data)

m.parameter_names = Set(initialize=[1, 2])

m.P = Var(m.parameter_names, bounds=(-5, 5), initialize=1)

m.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

@m.Objective(sense=minimize)
def objective(m):
    obj = 0
    for i in m.indx:
        obj += (m.P[1]*m.X[i] + m.P[2] - m.Y[i])**2
        
    return obj

""" Model 2 """

m2 = ConcreteModel()

x_data2 = dict(zip(list(range(1, 6)), list(range(1, 6))))
y_data2 = dict(zip(list(range(1, 6)), [1.50, 4.45, 6.47, 8.60, 10.96]))

m2.indx = Set(initialize=list(range(1, 6)))

m2.X = Param(m2.indx, initialize=x_data2)
m2.Y = Param(m2.indx, initialize=y_data2)

m2.parameter_names = Set(initialize=[1, 2])

m2.P = Var(m2.parameter_names, bounds=(-5, 5), initialize=1)

m2.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

@m2.Objective(sense=minimize)
def objective(m2):
    obj = 0
    for i in m2.indx:
        obj += (m2.P[1]*m2.X[i] + m2.P[2] - m2.Y[i])**2
        
    return obj

""" Model 3 """

m3 = ConcreteModel()

x_data3 = dict(zip(list(range(1, 6, 2)), list(range(1, 6, 2))))
y_data3 = dict(zip(list(range(1, 6, 2)), [4.67, 10.47, 15.23]))

m3.indx = Set(initialize=list(range(1, 6, 2)))

m3.X = Param(m3.indx, initialize=x_data3)
m3.Y = Param(m3.indx, initialize=y_data3)

m3.parameter_names = Set(initialize=[1, 2])

m3.P = Var(m3.parameter_names, bounds=(-5, 5), initialize=1)

m3.dual = Suffix(direction=Suffix.IMPORT_EXPORT)

@m3.Objective(sense=minimize)
def objective(m3):
    obj = 0
    for i in m3.indx:
        obj += (m3.P[1]*m3.X[i] + m3.P[2] - m3.Y[i])**2
        
    return obj

""" NSD Methods """

# Generate the ReactionModels (at some point KipetModel)
models = [m, m2, m3]

# Create the NSD object using the ReactionModels list
nsd = NSD(models, kwargs = dict(kipet=False))

# Set the initial values
nsd.set_initial_value({1 : 2,
                       2 : 1.2e-4}
                     )

# Runs the TR Method
#results = nsd.trust_region(scaled=False)
#nsd.plot_paths()

# Runs the Ipopt Method
results = nsd.ipopt_method(scaled=False)

for i, (k, v) in enumerate(m.P.items()):
    v.set_value(nsd.parameters_opt[k])

""" Plot the resulting line """

x = list(x_data.values())
y = list(y_data.values())
y2 = list(y_data2.values())
x3 = list(x_data3.values())
y3 = list(y_data3.values())

xm = np.linspace(1, 5, 9)
ym = xm*m.P[1].value + m.P[2].value

plt.plot(x, y, 'o')
plt.plot(x, y2, 'ro')
plt.plot(x3, y3, 'go')
plt.plot(xm, ym)


"""
Results from the various KIPET models are stored here
"""
# Standard library imports
import itertools
from pathlib import Path
import time

# Thirdparty library imports
import datetime
import numpy as np
import pandas as pd
from pyomo.core import *
from pyomo.core.base.PyomoModel import ConcreteModel
from pyomo.environ import *

# Kipet library imports
from kipet.post_model_build.pyomo_model_tools import (
    get_vars, 
    get_vars_block, 
    get_result_vars,
    get_index_sets,
    index_set_info,
    )         
from kipet.common.read_write_tools import df_from_pyomo_data

# This needs deletion
result_vars = ['Z', 'C', 'Cm', 'K', 'S', 'X', 'dZdt', 'dXdt', 'P', 'Pinit', 'sigma_sq', 'estimable_parameters', 'Y', 'UD', 'step']

class ResultsObject(object):
    """Container for all of the results. Includes plotting functions"""
    
    def __init__(self):
        """
        A class to store simulation and optimization results.
        """
        # Data series
        self.generated_datetime = datetime.datetime
        self.results_name = None
        self.solver_statistics = {}

    def __str__(self):
        string = "\nRESULTS\n"
        
        # results_vars = get_results_vars()
        
        for var in result_vars:
            if hasattr(self, var) and getattr(self, var) is not None:
                var_str = var
                if var == 'sigma_sq':
                    var_str = 'Sigmas2'
                string += f'{var_str}:\n {getattr(self, var)}\n\n'
        
        return string
    
    def __repr__(self):
        return self.__str__()

    def compute_var_norm(self, variable_name, norm_type=np.inf):
        var = getattr(self,variable_name)
        var_array = np.array(var)
        return np.linalg.norm(var_array,norm_type)
    
    @staticmethod
    def prepare_data_for_init(model, var):
        """Convert results dict into df for initialization
        This is used for data with dimensions larger than 2
        """
        #try:
        index_sets = get_index_sets(var)
        index_dict = index_set_info(index_sets)
        # print(var)
        # print(index_dict)
        time_set = index_sets[index_dict['cont_set'][0]].name
        component_indecies = index_dict['other_set']
        component_sets = [index_sets[i].name for i in component_indecies]
        index = getattr(model, time_set).value_list
        columns = list(itertools.product(*[getattr(model, comp_list).value_list for comp_list in component_sets]))
        df = pd.DataFrame(data=None, index=index, columns=columns)

        for i in index:
            for j in columns:
                jl = list(j)
                jl.insert(index_dict['cont_set'][0], i)
                df.loc[i,j] = var[tuple(jl)].value
                
        return df
        #except:
         #   return None

    def load_from_pyomo_model(self, instance, to_load=None):
        """Load variables from the pyomo model into various formats"""
        
        variables_to_load = get_vars(instance)
    
        for name in variables_to_load:
    
            if name == 'init_conditions':
                continue
            
            var = getattr(instance, name)
            if var.dim()==0:
                setattr(self, name, var.value)
            elif var.dim()==1:
                setattr(self, name, pd.Series(var.get_values()))
            elif var.dim()==2:
                d = var.get_values()
                keys = d.keys()
                if keys:
                    data_frame = df_from_pyomo_data(var)
                else:
                    data_frame = pd.DataFrame(data=[],
                                              columns = [],
                                              index=[])
                setattr(self, name, data_frame)        
            else:
                data_df = self.prepare_data_for_init(instance, var)
                setattr(self, name, data_df)
                
    @property
    def parameters(self):
        return self.P
            
    @property
    def show_parameters(self):
        print('\nThe estimated parameters are:')
        for k, v in self.P.items():
            print(k, v)
            
    @property
    def variances(self):
        print('\nThe estimated variances are:')
        for k, v in self.sigma_sq.items():
            print(k, v)


"""
initialization of dummy pyomo variables for use in model development

This should be updated to a more robust framework, eventually
"""
from pint import UnitRegistry
from pyomo.environ import ConcreteModel, Set, Var

class Component():
    """A generic pyomo variable to use in model building"""

    m = ConcreteModel()        
    m.indx = Set(initialize=[0])
    u = UnitRegistry()

    def __new__(cls,
                name,
                index=1,
                units=None):

        m = cls.m
        u = cls.u
        
        sets = [m.indx]*index
        m_units = u(units).units
        if m_units.dimensionless:
            m_units = str(1)
        else:
            m_units = getattr(u, str(m_units))
        
        setattr(m, name, Var(*sets, initialize=1, units=units))
        self = getattr(m, name)
        return self[tuple([0 for i in range(index)])]
    
class Parameter():
    """A `Symbol` representing a parameter whose value is known."""

    __qualname__ = 'Parameter'

    def __init__(self, 
                 name, 
                 value=0, 
                 index=1, 
                 units=None):
        
        self.name = name
        self.value = value
        self.index = index
        self.units = units
    
        self.Var = Component(name, index, units)

    # #def __new__(cls, name, value=0, index=1, units=None, parent=None):
    #     # self = Component.__new__(cls, name, index, units)
    #     # # self = super().__new__(cls, name)

    #     # self.value = value  # TODO: Add safeguards
    #     # self.units = units
    #     # return self

    # @property
    # def indexed(self):
    #     """Check whether the Parameter is indexed or not."""
    #     return self.expansion is not None

    # @property
    # def value(self):
    #     """Return the value or values of the Parameter."""
    #     return self._value

    # @value.setter
    # def value(self, data):
    #     """Set the value of the Parameter.

    #     In contrast to the Variable, a Parameter can be made indexed by
    #     simply providing some Mapping or a pandas.Series that imply
    #     both an index and values.
    #     """
    #     if isinstance(data, (Mapping, Series)):
    #         self.expand(data)
    #         self._value = None
    #     else:
    #         self._value = None if data is None else float(data)
    #         self.expansion = None
            
    # # @units.setter
    # # def units(self, data):
    # #     """Set the value of the Parameter.

    # #     In contrast to the Variable, a Parameter can be made indexed by
    # #     simply providing some Mapping or a pandas.Series that imply
    # #     both an index and values.
    # #     """
    # #     if data is not None:
    # #         self._units = cls.u(data)
    # #     else:
    # #         self._units = None
        
    # @property
    # def units(self):
    #     return self._units.units
        
    # @property
    # def indices(self): return self.expansion.keys()

    # @property
    # def elements(self): return self.expansion.values

    # @property
    # def items(self): return self.expansion.items()

    # @property
    # def parent(self):
    #     """Return the parent of this parameter."""
    #     return self._parent

    # def expand(self, data):
    #     """Expand the `Parameter` with indexed data."""
    #     if self._parent is not None:
    #         raise RuntimeError(f'Attempted to expand parameter {self} '
    #                            f'which is a member of {self._parent}!')
    #     self.expansion = Series((Parameter(f'{self.name}[{i}]', v,
    #                                        self)
    #                              for i, v in data.items()),
    #                             data.keys(), dtype='O')

    # def __getitem__(self, index):
    #     try:  # TODO: Custom errors still includes long stack trace...
    #         return self.expansion[index]
    #     except KeyError:
    #         if self.expansion is None:
    #             raise TypeError(f'Parameter {self.name} is not '
    #                             'indexed!')
    #         raise IndexError("Parameter index out of range")

    # def __setitem__(self, index, value):
    #     """Set the value of the element corresponding to the index."""
    #     self.expansion[index].value = value

    # def __iter__(self):
    #     """Iterate over the elements of this Parameter."""
    #     if self.expansion is None:
    #         raise TypeError(f'{self} is scalar.')
    #     for elem in self.elements:
    #         yield elem

    # @property
    # def is_Parameter(self):
    #     return True

if __name__ == '__main__':
    
    a = Parameter('k', value=1, units='1/s')
    b = Parameter('A', 10)


"""
Spectral Data Handling for Kipet
"""
import inspect
import os

import numpy as np
import pandas as pd
import plotly.graph_objs as go
from plotly import __version__
from plotly.offline import download_plotlyjs, init_notebook_mode, plot, iplot

from kipet.core_methods.data_tools import *
from kipet.visuals.plots import colors

class SpectralData():
    
    """All of the spectral tools will be moved here"""   
 
    def __init__(self, name, data=None, remove_negatives=False):
        
        self.name = name
        self.data = data
        self.data_orig = data
        self.remove_negatives=remove_negatives
        
        if self.remove_negatives and self.data is not None:
            self._remove_negatives()
        
        self._check_columns()
        
    def add_data(self, data):
        
        setattr(self, 'data', data)

        if self.remove_negatives:
            self._remove_negatives()
        if self.data_orig is None:
            setattr(self, 'data_orig', data)
            
        self._check_columns()
        
    def _check_columns(self):
        
        if hasattr(self, 'data') and self.data is not None:
            old_columns = self.data.columns
            new_columns = [float(col) for col in old_columns]
            self.data.columns = new_columns
        if hasattr(self, 'data_orig') and self.data_orig is not None:
            old_columns = self.data.columns
            new_columns = [float(col) for col in old_columns]
            self.data.columns = new_columns
            
    def reset(self):
        
        self.data = self.data_orig
        return None
        
    def plot(self, data_set='data'):
        """ Plots spectral data
        
            Args:
                dataFrame (DataFrame): spectral data
              
            Returns:
                None
    
        """
        data = getattr(self, data_set)
       
        fig = go.Figure()
        fig.add_trace(go.Surface(x=data.columns,
                           y=data.index,
                           z=data.values,
                          ))
        
        fig.update_layout(scene = dict(
                            xaxis_title='Wavelength',
                            yaxis_title='Time',
                            zaxis_title='Absorbance'),
                            margin=dict(r=100, b=50, l=100, t=50),
                            title_text=f'{self.name}: Spectral Data',
                            title_font_size=30)
        
        plot(fig)
            
        return None
    
    def _remove_negatives(self):
        
        self.data[self.data < 0] = 0
    
    def savitzky_golay(self, window_size=3, orderPoly=2, orderDeriv=0, in_place=True):
        """
        Implementation of the Savitzky-Golay filter for Kipet. Used for smoothing data, with
        the option to also differentiate the data. Can be used to remove high-frequency noise.
        Creates a least-squares fit of data within each time window with a high order polynomial centered
        centered at the middle of the window of points.
        
        Args:
            dataFrame (DataFrame): the data to be smoothed (either concentration or spectral data)
            window_size (int): the length of the window. Must be an odd integer number
            orderPoly (int): order of the polynoial used in the filter. Should be less than window_size-1
            orderDeriv (int) (optional): the order of the derivative to compute (default = 0 means only smoothing)
            
        Returns:
            DataFrame containing the smoothed data
        
        References:
            This code is an amalgamation of those developed in the scipy.org cookbook and that employed in Matlab 
            by WeiFeng Chen.
            Original paper: A. Savitzky, M. J. E. Golay, Smoothing and Differentiation of Data by 
            Simplified Least Squares Procedures. Analytical Chemistry, 1964, 36 (8), pp 1627-1639.
        """
        # data checks
        dataFrame = self.data
            #dataFrame = self.data_orig
        try:
            window_size = np.abs(np.int(window_size))
            orderPoly = np.abs(np.int(orderPoly))
        except ValueError:
            raise ValueError("window_size and order have to be of type int")
        if window_size % 2 != 1 or window_size < 1:
            raise TypeError("window_size size must be a positive odd number")
        if window_size < orderPoly + 2:
            raise TypeError("window_size is too small for the polynomials order")    
        if orderPoly >= window_size:
            raise ValueError("polyorder must be less than window_length.")
    
        if not isinstance(dataFrame, pd.DataFrame):
            raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
        print("Applying the Savitzky-Golay filter")
        
        order_range = range(orderPoly+1)
        half_window = (window_size -1) // 2
        # precompute coefficients
        b = np.mat([[k**i for i in order_range] for k in range(-half_window, half_window+1)])
        m = np.linalg.pinv(b).A[orderDeriv]
        #rate = 1
        #m = np.linalg.pinv(b).A[orderDeriv] * rate**orderDeriv * factorial(orderDeriv)
        D = np.array(dataFrame)
        no_noise = np.array(dataFrame)
        # pad the signal at the extremes with values taken from the signal itself
        for t in range(len(dataFrame.index)):
            row = list()
            for l in range(len(dataFrame.columns)):
                row.append(D[t,l])
            firstvals = row[0] - np.abs( row[1:half_window+1][::-1] - row[0] )
            lastvals = row[-1] + np.abs(row[-half_window-1:-1][::-1] - row[-1])
            y = np.concatenate((firstvals, row, lastvals))
            new_row = np.convolve( m, y, mode='valid')
            no_noise[t]=new_row
            
        if orderDeriv == 0:
            for t in range(len(dataFrame.index)):
                for l in range(len(dataFrame.columns)):
                    if no_noise[t,l] < 0:
                        no_noise[t,l] = 0
        
        data_frame = pd.DataFrame(data=no_noise,
                                  columns = dataFrame.columns,
                                  index=dataFrame.index)
        
        if in_place:
            self.data = data_frame
        
        return data_frame
    
    def snv(self, offset=0, in_place=True):
        """
        Implementation of the Standard Normal Variate (SNV) filter for Kipet which is a weighted normalization
        method that is commonly used to remove scatter effects in spectroscopic data, this pre-processing 
        step can be applied before the SG filter or used on its own. SNV can be sensitive to noisy entries 
        in the spectra and can increase nonlinear behaviour between S and C as it is not a linear transformation.
        
        
        Args:
            dataFrame (DataFrame): the data to be processed (either concentration or spectral data)
            offset (float): user-defined offset which can be used to avoid over-normalization for samples
                            with near-zero standard deviation. Guide for choosing this value is for something 
                            near the expected noise level to be specified. Default value is zero.
            
        Returns:
            DataFrame containing pre-processed data
        
        References:
    
        """
        dataFrame = self.data
        # if dataFrame is None:
        #     dataFrame = self.data_orig
            
        # data checks
        if not isinstance(dataFrame, pd.DataFrame):
            raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
        print("Applying the SNV pre-processing")    
    
        D = np.array(dataFrame)
        snv_proc = np.array(dataFrame)
        for t in range(len(dataFrame.index)):
            row = list()
            sum_spectra = 0
            for l in range(len(dataFrame.columns)):
                row.append(D[t,l])
                sum_spectra += D[t,l]
            mean_spectra = sum_spectra/(len(dataFrame.columns))
            std = 0
            for l in range(len(dataFrame.columns)):
                std += (mean_spectra-D[t,l])**2
            new_row = list()
            for l in range(len(dataFrame.columns)):
                if offset ==0:
                    w = (D[t,l]-mean_spectra)*(std/(len(dataFrame.columns)-1))**0.5
                else:
                    w = (D[t,l]-mean_spectra)*(std/(len(dataFrame.columns)-1))**0.5 + 1/offset
                new_row.append(w)
                    
            snv_proc[t]=new_row
    
        data_frame = pd.DataFrame(data=snv_proc,
                                  columns = dataFrame.columns,
                                  index=dataFrame.index)
        if in_place:
            self.data = data_frame
        return data_frame
    
    def msc(self, reference_spectra=None, in_place=True):
        """
        Implementation of the Multiplicative Scatter Correction (MSC) filter for Kipet which is simple pre-processing
        method that attempts to remove scaling effects and offset effects in spectroscopic data. This pre-processing 
        step can be applied before the SG filter or used on its own. This approach requires a reference spectrum which
        must be determined beforehand. In this implementation, the default reference spectrum is the average spectrum 
        of the dataset provided, however an optional argument exists for user-defined reference spectra to be provided.    
        
        Args:
            dataFrame (DataFrame):          the data to be processed (either concentration or spectral data)
            reference_spectra (DataFrame):  optional user-provided reference spectra argument. Default is to automatically
                                            determine this using the average spectra values.
            
        Returns:
            DataFrame pre-processed data
        
        References:
    
        """
        dataFrame = self.data
        # if dataFrame is None:
        #     dataFrame = self.data_orig
        # data checks
        if not isinstance(dataFrame, pd.DataFrame):
            raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
        print("Applying the MSC pre-processing")  
        
        #Want to make it possible to include user-defined reference spectra
        #this is not great as we could provide the data with some conditioning 
        #in order to construct references based on different user inputs
        if reference_spectra != None:
            if not isinstance(reference_spectra, pd.DataFrame):
                raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
            
            if len(dataFrame.columns) != len(reference_spectra.columns) and len(dataFrame.rows) != len(reference_spectra.rows):
                raise NotImplementedError("the reference spectra must have the same number of entries as the data")
        
        D = np.array(dataFrame)
        ref = np.array(dataFrame)
        msc_proc = np.array(dataFrame)
        
        # the average spectrum is calculated as reference spectra for MSC when none is given by user
        if reference_spectra == None:
            sum_spectra = 0
            
            for t in range(len(dataFrame.index)):
                sum_spectra = 0
                for l in range(len(dataFrame.columns)):
                    sum_spectra += D[t,l]
                mean_spectra = sum_spectra/(len(dataFrame.columns))
                for l in range(len(dataFrame.columns)):
                    ref[t,l] = mean_spectra 
        else:
            #should add in some checks and additional ways to formulate these depending on what input the user provides
            #need to find out the type of data usually inputted here in order to do this
            ref = reference_spectra
        for t in range(len(dataFrame.index)):
            row = list()
            fit = np.polyfit(ref[t,:],D[t,:],1, full=True)
            row[:] = (D[t,:] - fit[0][1]) / fit[0][0]
            msc_proc[t,:]=row  
    
        data_frame = pd.DataFrame(data=msc_proc,
                                  columns = dataFrame.columns,
                                  index=dataFrame.index)
        if in_place:
            self.data = data_frame
        return data_frame
    
    def baseline_shift(self, shift=None, in_place=True):
        """
        Implementation of basic baseline shift. 2 modes are avaliable: 1. Automatic mode that requires no
        user arguments. The method identifies the lowest value (NOTE THAT THIS ONLY WORKS IF LOWEST VALUE
        IS NEGATIVE) and shifts the spectra up until this value is at zero. 2. Baseline shift provided by
        user. User provides the number that is added to every wavelength value in the full spectral dataset.
        
        
        Args:
            dataFrame (DataFrame): the data to be processed (spectral data)
            shift (float, optional): user-defined baseline shift
            
        Returns:
            DataFrame containing pre-processed data
        
        References:
    
        """
        dataFrame = self.data
        # if dataFrame is None:
        #     dataFrame = self.data_orig
        # data checks
        if not isinstance(dataFrame, pd.DataFrame):
            raise TypeError("data must be inputted as a pandas DataFrame, try using read_spectral_data_from_txt or similar function first")
        print("Applying the baseline shift pre-processing") 
        if shift == None:
            shift = float(dataFrame.min().min())*(-1)
        
        print("shifting dataset by: ", shift)    
        D = np.array(dataFrame)
        for t in range(len(dataFrame.index)):
            for l in range(len(dataFrame.columns)):
                D[t,l] = D[t,l]+shift
        
        data_frame = pd.DataFrame(data=D, columns = dataFrame.columns, index = dataFrame.index)
        
        if in_place:
            self.data = data_frame
        return data_frame
    
    def decrease_wavelengths(self, A_set=2, specific_subset=None, in_place=True):
        '''
        Takes in the original, full dataset and removes specific wavelengths, or only keeps every
        multiple of A_set. Returns a new, smaller dataset that should be easier to solve
        
        Args:
            original_dataset (DataFrame):   the data to be processed
            A_set (float, optional):  optional user-provided multiple of wavelengths to keep. i.e. if
                                        3, every third value is kept. Default is 2.
            specific_subset (list or dict, optional): If the user already knows which wavelengths they would like to
                                        remove, then a list containing these can be included.
            
        Returns:
            DataFrame with the smaller dataset
        
        '''
        original_dataset = self.data
        # if original_dataset is None:
        #     original_dataset = self.data_orig
            
        if specific_subset != None:
            if not isinstance(specific_subset, (list, dict)):
                raise RuntimeError("subset must be of type list or dict!")
                 
            if isinstance(specific_subset, dict):
                lists1 = sorted(specific_subset.items())
                x1, y1 = zip(*lists1)
                specific_subset = list(x1)
                
            new_D = pd.DataFrame(np.nan,index=original_dataset.index, columns = specific_subset)
            for t in original_dataset.index:
                for l in original_dataset.columns.values:
                    if l in subset:
                        new_D.at[t,l] = self.model.D[t,l]           
        else:
            count=0
            for l in original_dataset.columns.values:
                remcount = count%A_set
                if remcount==0:
                    original_dataset.drop(columns=[l],axis = 1)
                count+=1
            new_D = original_dataset[original_dataset.columns[::A_set]] 
            
        if in_place:
            self.data = new_D
        return new_D


"""
Replacement and Scaling Mixins used in model modification
"""
from pyomo.core.expr import current as EXPR
from pyomo.environ import *
from pyomo.core.expr.numvalue import NumericConstant

class ReplacementVisitor(EXPR.ExpressionReplacementVisitor):
    
    def __init__(self):
        super(ReplacementVisitor, self).__init__()
        self._replacement = None
        self._suspect = None

    def change_suspect(self, suspect_):
        self._suspect = suspect_

    def change_replacement(self, replacement_):
        self._replacement = replacement_

    def visiting_potential_leaf(self, node):
        #
        # Clone leaf nodes in the expression tree
        #
        if node.__class__ in native_numeric_types:
            return True, node

        if node.__class__ is NumericConstant:
            return True, node


        if node.is_variable_type():
            if id(node) == self._suspect:
                d = self._replacement
                return True, d
            else:
                return True, node
            
        if node.is_parameter_type():
            if id(node) == self._suspect:
                d = self._replacement
                return True, d
            else:
                return True, node

        return False, None
    
class ScalingVisitor(EXPR.ExpressionReplacementVisitor):

    def __init__(self, scale):
        super(ScalingVisitor, self).__init__()
        self.scale = scale

    def visiting_potential_leaf(self, node):
      
        if node.__class__ in native_numeric_types:
            return True, node

        if node.is_variable_type():
            #print(node)
            return True, self.scale[id(node)]*node

        if isinstance(node, EXPR.LinearExpression):
            node_ = copy.deepcopy(node)
            node_.constant = node.constant
            node_.linear_vars = copy.copy(node.linear_vars)
            node_.linear_coefs = []
            for i,v in enumerate(node.linear_vars):
                node_.linear_coefs.append( node.linear_coefs[i]*self.scale[id(v)] )
            return True, node_

        return False, None


class FindingVisitor(EXPR.ExpressionValueVisitor):
            
    def __init__(self):
        super(FindingVisitor, self).__init__()
        self._suspect = None
        self._found = False

    def find_suspect(self, suspect_):
        self._suspect = suspect_

    def visiting_potential_leaf(self, node):
        """Clone leaf in expression tree
        
        """
        if id(node) == self._suspect:
             self._found = True
  
        if node.__class__ in native_numeric_types:
            return True, node

        if node.__class__ is NumericConstant:
            return True, node

        if node.is_variable_type():
            return True, node
            
        if node.is_parameter_type():
            return True, node
            
        return False, None


"""
Read/Write functions used for data in Kipet
"""
from contextlib import contextmanager, redirect_stdout
import inspect
import os
from pathlib import Path
import re
import sys

import numpy as np
import pandas as pd

#from kipet.top_level.settings import USER_DEFINED_SETTINGS

DEFAULT_DIR = Path.cwd()

# def set_directory(filename, data_dir=DEFAULT_DIR, abs_dir=False):
#     """Sets the current working directory plus the given subdirectory as the
#     directory for the given data file (filename)
    
#     Args:
#         filename (str): name of input file
        
#         directory (str): name of directory with data file
          
#     Returns:
#         filename (Path): the full filename of the data
    
#     """
#     if not abs_dir:
#         # This should be a relative path to the calling file
#         calling_file_name = os.path.dirname(os.path.realpath(sys.argv[0]))
#         print(calling_file_name)
#         dataDirectory = Path(calling_file_name).joinpath(data_dir, filename)
#         return dataDirectory

#     else:
#         # It uses the absolute path provided
#         return Path(filename)

def read_file(filename, directory=DEFAULT_DIR):       
    """ Reads data from a csv or txt file and converts it to a DataFrame
    
        Args:
            filename (str): name of input file (abs path)
          
        Returns:
            DataFrame

    """
    
    filename = Path(filename)
    data_dict = {}
    
    if filename.suffix == '.txt':
        with open(filename, 'r') as f:
            for line in f:
                if line not in ['','\n','\t','\t\n']:
                    l = line.split()
                    if is_float_re(l[1]):
                        l[1] = float(l[1])
                    data_dict[float(l[0]), l[1]] = float(l[2])
        
        df_data = dict_to_df(data_dict)
        df_data.sort_index(ascending=True, inplace=True)

    elif filename.suffix == '.csv':
        df_data = pd.read_csv(filename, index_col=0)   

    else:
        raise ValueError(f'The file extension {filename.suffix} is currently not supported')
        return None

    return df_data
    

def write_file(filename, dataframe, filetype='csv'):
    """ Write data to file.
    
        Args:
            filename (str): name of output file (abs_path)
          
            dataframe (DataFrame): pandas DataFrame
        
            filetype (str): choice of output (csv, txt)
        
        Returns:
            None

    """
    # How can you write a general settings file/class/object?
    print(f'Here is the filename: {filename}')
    
    if filetype not in ['csv', 'txt']:
        print('Savings as CSV - invalid file extension given')
        filetype = 'csv'
    
    suffix = '.' + filetype
    
    filename = Path(filename)
                   
    if filename.suffix == '':
        filename = filename.with_suffix(suffix)
    else:
        suffix = filename.suffix
        if suffix not in ['.txt', '.csv']:
            print('Savings as CSV - invalid file extension given')
            filename = Path(filename.stem).with_suffix('.csv')
    
    print(f'In write method: {filename.absolute()}')
    if filename.suffix == '.csv':
        dataframe.to_csv(filename)

    elif filename.suffix == '.txt':    
        with open(filename, 'w') as f:
            for i in dataframe.index:
                for j in dataframe.columns:
                    if not np.isnan(dataframe[j][i]):
                        f.write("{0} {1} {2}\n".format(i,j,dataframe[j][i]))
                        
    print(f'Data saved     : {filename}')
    return None

def read_spectral_data_from_csv(filename, instrument = False, negatives_to_zero = False):
    """ Reads csv with spectral data
    
        Args:
            filename (str): name of input file
            instrument (bool): if data is direct from instrument
            negatives_to_zero (bool): if data contains negatives and baseline shift is not
                                        done then this forces negative values to zero.

        Returns:
            DataFrame

    """
    data = pd.read_csv(filename, index_col=0)
    if instrument:
        #this means we probably have a date/timestamp on the columns
        data = pd.read_csv(filename,index_col=0, parse_dates = True)
        data = data.T
        for n in data.index:
            h,m,s = n.split(':')
            sec = (float(h)*60+float(m))*60+float(s)
            data.rename(index={n:sec}, inplace=True)
        data.index = [float(n) for n in data.index]
    else:
        data.columns = [float(n) for n in data.columns]

    #If we have negative values then this makes them equal to zero
    if negatives_to_zero:
        data[data < 0] = 0

    return data

# Legacy functions

# def write_spectral_data_to_csv(filename, dataframe, directory=DEFAULT_DIR):
#     """
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
#     """
#     write_file(filename, dataframe, 'csv', directory)

# def write_spectral_data_to_txt(filename, dataframe, directory=DEFAULT_DIR):
#     """ Write spectral data Dij to txt file.
    
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
    
#     """
#     write_file(filename, dataframe, 'txt', directory)

# def write_absorption_data_to_csv(filename, dataframe, directory=DEFAULT_DIR):
#     """ Write absorption data Sij to csv file.
    
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
#     """
#     write_file(filename, dataframe, 'csv', directory)

# def write_absorption_data_to_txt(filename, dataframe, directory=DEFAULT_DIR):
#     """ Write absorption data Sij to txt file.
    
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
#     """
#     write_file(filename, dataframe, 'txt', directory)

# def write_concentration_data_to_csv(filename, dataframe, directory=DEFAULT_DIR):
#     """ Write concentration data Cij to csv file.
    
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
#     """
#     write_file(filename, dataframe, 'csv', directory)

# def write_concentration_data_to_txt(filename, dataframe, directory=DEFAULT_DIR):
#     """ Write concentration data Cij to txt file.
    
#         Args:
#             filename (str): name of output file
          
#             dataframe (DataFrame): pandas DataFrame
        
#         Returns:
#             None
#     """
#     write_file(filename, dataframe, 'txt', directory)


# def read_concentration_data(filename):
    
#     print(filename)
#     if filename.suffix == '.csv':
#         return read_concentration_data_from_csv(filename)
#     elif filename.suffix == '.txt':
#         return read_concentration_data_from_txt(filename)
#     else:
#         raise ValueError('Filetype not csv or txt.')
#         return None

# def read_concentration_data_from_txt(filename, directory=DEFAULT_DIR):
#     """ Reads txt with concentration data
    
#         Args:
#             filename (str): name of input file
          
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory)

# def read_concentration_data_from_csv(filename, directory=DEFAULT_DIR):
#     """ Reads csv with concentration data
    
#         Args:
#             filename (str): name of input file
         
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory)


# def read_absorption_data_from_csv(filename, directory=DEFAULT_DIR):
#     """ Reads csv with spectral data
    
#         Args:
#             filename (str): name of input file
          
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory)


# def read_spectral_data_from_txt(filename, directory=DEFAULT_DIR, **kwargs):
#     """ Reads txt with spectral data
    
#         Args:
#             filename (str): name of input file
          
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory)

# def read_spectral_data_from_csv(filename, directory=DEFAULT_DIR, **kwargs):
#     """ Reads txt with spectral data
    
#         Args:
#             filename (str): name of input file
          
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory, **kwargs)


# def read_absorption_data_from_txt(filename, directory=DEFAULT_DIR):
#     """ Reads txt with absorption data
    
#         Args:
#             filename (str): name of input file
          
#         Returns:
#             DataFrame
#     """
#     return read_file(filename, directory)


# for redirecting stdout to files
@contextmanager
def stdout_redirector(stream):
    old_stdout = sys.stdout
    sys.stdout = stream
    try:
        yield
    finally:
        sys.stdout = old_stdout
    
# Data conversion tools

def dict_to_df(data_dict):

    """Takes a dictionary of typical pyomo data and converts it to a dataframe
    
    """    
    dfs_stacked = pd.Series(index=data_dict.keys(), data=list(data_dict.values()))
    dfs = dfs_stacked.unstack()
    return dfs

def is_float_re(str):
    """Checks if a value is a float or not"""
    _float_regexp = re.compile(r"^[-+]?(?:\b[0-9]+(?:\.[0-9]*)?|\.[0-9]+\b)(?:[eE][-+]?[0-9]+\b)?$").match
    return True if _float_regexp(str) else False

def df_from_pyomo_data(varobject):

    val = []
    ix = []
    for index in varobject:
        ix.append(index)
        val_raw = varobject[index].value
        if val_raw is None:
            val_raw = 0
        val.append(val_raw)
    
    a = pd.Series(index=ix, data=val)
    dfs = pd.DataFrame(a)
    index = pd.MultiIndex.from_tuples(dfs.index)
   
    dfs = dfs.reindex(index)
    dfs = dfs.unstack()
    dfs.columns = [v[1] for v in dfs.columns]

    return dfs



import bisect
import math

import numpy as np
import scipy

from pyomo.dae import *
from pyomo.environ import *
from kipet.core_methods.ResultsObject import *

def interpolate_trajectory(t, tr):
    
    tr_val = np.zeros((len(t)))
    times = [float(ti) for ti in t]
    indx = tr.index.astype(float)

    for i, t_indx in enumerate(times):

        if i == 0:
            tr_val[i] = tr.iloc[0]
            continue        

        indx_left = bisect.bisect_left(indx[1:], times[i])
        if indx_left == len(tr) - 1:
            tr_val[i] = tr.iloc[indx_left]
            continue
        
        dx = indx[indx_left + 1] - indx[indx_left]
        dy = tr.iloc[indx_left + 1] - tr.iloc[indx_left]
        m = dy/dx
        val = tr.iloc[indx_left] +  (times[i] - indx[indx_left]) * m
        tr_val[i] = val
    
    return tr_val 

class Simulator(object):
    """Base simulator class.

    Note:
        This class is not intended to be used directly by users

    Attributes:
        model (model): Casadi or Pyomo model.

    """

    def __init__(self,model):
        self.model = model

        # make model atributes of the model global.
        # Need to be remove, and just get things from the model directly

        self._mixture_components = [name for name in self.model.mixture_components]
        self._complementary_states = [name for name in self.model.complementary_states]
        self._algebraics = [name for name in self.model.algebraics]
        self._meas_times = sorted([t for t in self.model.meas_times])
        self._allmeas_times = sorted([t for t in self.model.allmeas_times])#added for new data structure CS
        self._meas_lambdas = sorted([l for l in self.model.meas_lambdas])
        self._n_meas_times = len(self._meas_times)
        self._n_allmeas_times = len(self._allmeas_times)#added for new data structure CS
        self._n_meas_lambdas = len(self._meas_lambdas)
        self._n_components = len(self._mixture_components)
        self._n_algebraics = len(self._algebraics)
        self._n_complementary_states = len(self._complementary_states)
        self._non_absorbing = None
        self._huplc_absorbing = None #added for additional huplc data CS
        self._known_absorbance = None
        self._known_absorbance_data = None
        
        if hasattr(self.model, 'non_absorbing'):
            self._non_absorbing = [name for name in self.model.non_absorbing]
        if hasattr(self.model, 'known_absorbance'):
            self._known_absorbance = [name for name in self.model.known_absorbance]
        #added for removing nonabs ones from first term in objective CS:
        if hasattr(self.model, 'abs_components'):
            self._abs_components = [name for name in
                                    self.model.abs_components]  # added for removing nonabs ones from first term in objective CS
            self._nabs_components = len(self._abs_components)
        ####
        #added for additional huplc data CS:
        if hasattr(self.model, 'huplc_absorbing'):
            self._huplcmeas_times = sorted([t for t in self.model.huplcmeas_times])  # added for additional huplc data CS
            self._n_huplcmeas_times = len(self._huplcmeas_times)  # added for additional huplc data CS
            self._huplc_absorbing = [name for name in self.model.huplc_absorbing]
        if hasattr(self.model, 'huplcabs_components'):
            self._huplcabs_components = [name for name in
                                        self.model.huplcabs_components]
            self._nhuplcabs_components = len(self._huplcabs_components)
        ####
        if not self._mixture_components:
            raise RuntimeError('The model does not have any mixture components.\
            For simulation add mixture components')

    def apply_discretization(self, transformation, **kwargs):
        raise NotImplementedError("Simulator abstract method. Call child class")

    def initialize_from_trajectory(self,trajectory_dictionary):
        raise NotImplementedError("Simulator abstract method. Call child class")
    
    def run_sim(self,solver,**kwds):
        raise NotImplementedError("Simulator abstract method. Call child class")

    def compute_D_given_SC(self,results,sigma_d=0):
        # this requires results to have S and C computed already
        d_results = []

        if hasattr(self, '_abs_components'):  # added for removing non_abs ones from first term in obj CS
            for i, t in enumerate(self._allmeas_times):
                if t in self._meas_times:
                    for j, l in enumerate(self._meas_lambdas):
                        suma = 0.0
                        for w, k in enumerate(self._abs_components):
                            Cs = results.Cs[k][t]  # just the absorbing ones
                            Ss = results.S[k][l]
                            suma += Cs * Ss
                        if sigma_d:
                            suma += np.random.normal(0.0, sigma_d)
                        d_results.append(suma)

        else:
            for i, t in enumerate(self._allmeas_times):
                if t in self._meas_times:
                    for j, l in enumerate(self._meas_lambdas):
                        suma = 0.0
                        for w, k in enumerate(self._mixture_components):
                            C = results.C[k][t]
                            S = results.S[k][l]
                            suma += C * S
                        if sigma_d:
                            suma += np.random.normal(0.0, sigma_d)
                        d_results.append(suma)

        d_array = np.array(d_results).reshape((self._n_meas_times, self._n_meas_lambdas))
        results.D = pd.DataFrame(data=d_array,
                                 columns=self._meas_lambdas,
                                 index=self._meas_times)